#
#include	"dab-constants.h"
#include	"dab_concurrent.h"
#include	<QThread>
#include	<QMutex>
#include	<QWaitCondition>
#ifdef	KJMP
#include	"mp2processor-kjmp.h"
#else
#include	"mp2processor-ffmpeg.h"
#endif
#include	"mp4processor.h"
#include	"deconvolve.h"
#include	"audiosink.h"
#include	"gui.h"
//
//	As an experiment a version of the backend is created
//	that will be running in a separate thread. Might be
//	useful for multicore processors.
//
//	Interleaving is - for reasons of simplicity - done
//	inline rather than through a special class-object
static
int8_t	interleaveDelays [] = {
	     15, 7, 11, 3, 13, 5, 9, 1, 14, 6, 10, 2, 12, 4, 8, 0};
//
//
//	fragmentsize == Length * CUSize
	dabConcurrent::dabConcurrent	(uint8_t dabModus,
	                                 int16_t fragmentSize,
	                                 int16_t bitRate,
	                                 int16_t uepFlag,
	                                 int16_t protLevel,
	                                 RadioInterface *mr,
	                                 FILE	*mp2file,
	                                 FILE	*mp4file,
	                                 audioSink *as) {
int32_t i, j;
	this	-> dabModus		= dabModus;
	this	-> fragmentSize		= fragmentSize;
	this	-> bitRate		= bitRate;
	this	-> uepFlag		= uepFlag;
	this	-> protLevel		= protLevel;
	this	-> myRadioInterface	= mr;
	this	-> mp2File		= mp2file;
	this	-> mp4File		= mp4file;
	this	-> myAudioSink		= as;

	outV			= new uint8_t [bitRate * 24];
	interleaveData		= new int16_t *[fragmentSize]; // max size
	Data			= new int16_t [fragmentSize];
	for (i = 0; i < fragmentSize; i ++) {
	   interleaveData [i] = new int16_t [16];
	   for (j = 0; j < 16; j ++)
	      interleaveData [i][j] = 0;
	}

	uepProcessor		= NULL;
	eepProcessor		= NULL;
	if (uepFlag == 0)
	   uepProcessor	= new uep_deconvolve (bitRate,
	                                      protLevel);
	else
	   eepProcessor	= new eep_deconvolve (bitRate,
	                                      protLevel);
//
	
	if (dabModus == DAB) 
	   our_dabProcessor = new mp2Processor (myRadioInterface,
	                                        myAudioSink,
	                                        mp2File,
	                                        bitRate);
	else
	if (dabModus == DAB_PLUS) 
	   our_dabProcessor = new mp4Processor (myRadioInterface,
	                                        myAudioSink,
	                                        mp4File,
	                                        bitRate);
	else		// cannot happen
	   our_dabProcessor = new dabProcessor ();

	myAudioSink	-> restart	();
	Buffer		= new RingBuffer<int16_t>(64 * 32768);
	running		= true;
	start ();
}

	dabConcurrent::~dabConcurrent	(void) {
int16_t	i;
	running = false;
	while (this -> isRunning ())
	   usleep (1);
	if (uepProcessor != NULL)
	   delete uepProcessor;
	if (eepProcessor != NULL)
	   delete eepProcessor;
	delete our_dabProcessor;
	delete	Buffer;
	delete[]	outV;
	for (i = 0; i < fragmentSize; i ++) 
	   delete[]  interleaveData [i];
	delete [] interleaveData;
	delete [] Data;
}

int32_t	dabConcurrent::process	(int16_t *v, int16_t cnt) {
int32_t	fr;
	   while ((fr = Buffer -> GetRingBufferWriteAvailable ()) <= cnt) {
	      if (!running)
	         return 0;
	      usleep (1);
	   }

	   Buffer	-> putDataIntoBuffer (v, cnt);
	   Locker. wakeAll ();
	   return fr;
}

void	dabConcurrent::run	(void) {
int16_t	i, j;
int32_t	countforInterleaver	= 0;
uint8_t	shiftRegister [9];

	while (running) {
	   while (Buffer -> GetRingBufferReadAvailable () <= fragmentSize) {
	      ourMutex. lock ();
	      Locker. wait (&ourMutex, 1);	// 1 msec waiting time
	      ourMutex. unlock ();
	      if (!running)
	         break;
	   }

	   if (!running) 
	      break;

	   Buffer	-> getDataFromBuffer (Data, fragmentSize);
	   for (i = 0; i < fragmentSize; i ++) {
	      interleaveData [i][interleaveDelays [i & 017]] = Data [i];
	      Data [i] = interleaveData [i] [0];
//	and shift
	      memmove (&interleaveData [i][0],
	               &interleaveData [i][1],
	               interleaveDelays [i & 017] * sizeof (int16_t));
	   }
//
//	only continue when de-interleaver is filled
	   if (countforInterleaver <= 15) {
	      countforInterleaver ++;
	      continue;
	   }
//
	   if (uepFlag == 0)
	      uepProcessor -> deconvolve (Data, fragmentSize, outV);
	   else
	      eepProcessor -> deconvolve (Data, fragmentSize, outV);
//
//	and the inline energy dispersal
	   memset (shiftRegister, 1, 9);
	   for (i = 0; i < bitRate * 24; i ++) {
	      uint8_t b = shiftRegister [8] ^ shiftRegister [4];
	      for (j = 8; j > 0; j--)
	         shiftRegister [j] = shiftRegister [j - 1];
	      shiftRegister [0] = b;
	      outV [i] ^= b;
	   }
	   our_dabProcessor -> addtoFrame (outV, 24 * bitRate);
	}
}
//
//	It might take a msec for the task to stop
void	dabConcurrent::stopRunning (void) {
	running = false;
	while (this -> isRunning ())
	   usleep (1);
//	myAudioSink	-> stop ();
}

void	dabConcurrent::setFiles (FILE *f1, FILE *f2) {
	mp2File		= f1;
	mp4File		= f2;
	if (dabModus == DAB)
	   our_dabProcessor	-> setFile (f1);
	else
	   our_dabProcessor	-> setFile (f2);
}

