#
#include	"mp2processor-ffmpeg.h"
#include	"limits.h"
#include	<cstring>
#include	"gui.h"

#define		INPUTBUFFER_SIZE	5120
//
//	
// sample rate table
static const unsigned short sample_rates[8] = {
    44100, 48000, 32000, 0,  // MPEG-1
    22050, 24000, 16000, 0   // MPEG-2
};

int32_t	mp2sampleRate	(uint8_t *frame) {
    if (!frame)
        return 0;
    if (( frame[0]         != 0xFF)   // no valid syncword?
    ||  ((frame[1] & 0xF6) != 0xF4)   // no MPEG-1/2 Audio Layer II?
    ||  ((frame[2] - 0x10) >= 0xE0))  // invalid bitrate?
        return 0;
    return sample_rates[(((frame[1] & 0x08) >> 1) ^ 4)  // MPEG-1/2 switch
                      + ((frame[2] >> 2) & 3)];         // actual rate
}

#define	SAMPLES_PER_FRAME	1152

	mp2Processor::mp2Processor (RadioInterface *mr,
	                            audioSink * as,
	                            FILE *mp2file,
	                            int16_t bitRate) {

	myRadioInterface	= mr;
	ourSink		= as;
	mp2File		= mp2file;
	connect (this, SIGNAL (show_successRate (int)),
	         mr, SLOT (show_successRate (int)));
	processorOK	= false;
	MP2fillState	= 0;
	MP2frame	= new uint8_t [INPUTBUFFER_SIZE + 100];
	MP2headerCount	= 0;
	MP2framesize	= 24 * bitRate;
	numberofFrames	= 0;
	numberofErrors	= 0;

	avcodec_register_all	();
	av_init_packet		(&avpkt);
	codec		= avcodec_find_decoder (CODEC_ID_MP1);
	if (codec == NULL) {
	   fprintf (stderr, "Codec not found\n");
	   return;
	}

	context		= avcodec_alloc_context3 (codec);
	if (context == NULL) {
	   fprintf (stderr, "Could not allocate codec context\n");
	   return;
	}

	if (avcodec_open2 (context, codec, NULL) < 0) {
	   fprintf (stderr, "Could not open codec\n");
	   return;
	}

	decoded_frame	= avcodec_alloc_frame ();
	if (decoded_frame == NULL) {
	   fprintf (stderr, "Could not allocate audioframe\n");
	   return;
	}

	baudRate	= 48000;	// default for DAB
//	ourSink		-> setMode (0, baudRate);
	processorOK	= true;
}

	mp2Processor::~mp2Processor	(void) {
	av_free (context);
}
//
void	mp2Processor::processCurrentFrame (int16_t MP2framebytes) {
int16_t	i;
int16_t	len;
int32_t	got_frame;

	numberofFrames ++;
	if (numberofFrames >= 50) {
	   show_successRate (100 - 2 * numberofErrors);
	   numberofFrames 	= 0;
	   numberofErrors	= 0;
	}

//	we only send frames to this function with a regular first 12 bits
//	so, here we check MPEG1 layer II
	if ((MP2frame [1] & 0x6) != 0x4) { // MPEG1 layer II
	   numberofErrors ++;
	   return;
	}

	for (i = MP2framebytes; i < MP2framebytes + 20; i ++)
	   MP2frame [i] = 0;

	if (mp2File != NULL) {
	   (void)fwrite (MP2frame, sizeof (uint8_t), MP2framebytes, mp2File);
	   return;
	}
	avcodec_get_frame_defaults (decoded_frame);

	avpkt. data	= MP2frame;
	avpkt. size	= MP2framebytes + 20;
	len	= avcodec_decode_audio4 (context,
	                                 decoded_frame,
	                                 &got_frame,
	                                 &avpkt);

	if (len < 0) {	// an error 
	   numberofErrors ++;
	   return;
	}

	if (got_frame) {
	   int32_t bs =  av_samples_get_buffer_size (NULL,
	                                       context -> channels,
	                                       decoded_frame -> nb_samples,
	                                       context -> sample_fmt, 1);
	   int32_t sample_rate	= context -> sample_rate;
	   int16_t amount	= decoded_frame -> nb_samples;
	   int16_t channels	= context -> channels;

	   if (channels == 2)
	      ourSink	-> audioOut ((int16_t *)(decoded_frame -> data [0]),
	                             amount, sample_rate);
	   else
	   if (channels == 1) {
	      int16_t *buffer = (int16_t *)alloca (2 * bs);
	      int16_t i;
	      for (i = 0; i < amount; i ++) {
	         buffer [2 * i]	= ((int16_t *)(decoded_frame -> data [0])) [i];
	         buffer [2 * i + 1] = buffer [2 * i];
	      }
	      ourSink -> audioOut (buffer, amount, sample_rate);
	   }
	}
	else
	   fprintf (stderr, "We did not get a frame\n");
}
//
//	It is a little annoying that the outputrate (either 48000 or
//	24000) is encoded, so that we have to process part of the
//	MP2 frame to find out what to do.
//	MP2Head
void	mp2Processor::addtoFrame (uint8_t *v, int16_t amount) {
int16_t	i, j;
int16_t x;
int16_t	lf	= baudRate == 48000 ? MP2framesize : 2 * MP2framesize;

	for (i = 0; i < amount; i ++) {
	   switch (MP2fillState) {
	      default:
	      case 0:		// look for sync
	         if (v [i] == 01) {
	            if (++ MP2headerCount == 12) {
	               MP2bitCount = 0;
	               for (j = 0; j < 12; j ++)
	                  addbittoMP2 (MP2frame, 1, MP2bitCount ++);
	               MP2fillState = 1;
	            }
	         }
	         else
	            MP2headerCount = 0;
	         break;

	      case 1:		// look for samplerate
	         addbittoMP2 (MP2frame, v [i], MP2bitCount ++);
	         if (MP2bitCount == 24) {	// 3 bytes read
	            int32_t newRate = mp2sampleRate (MP2frame);
	            if (newRate == 0) {
	               MP2fillState = 0;
	               MP2bitCount  = 0;
	               break;
	            }
	            if (newRate != baudRate) {
	               baudRate = newRate;
	               lf = baudRate == 48000 ?
	                              MP2framesize : 2 * MP2framesize;
	            }
	            MP2fillState = 2;
	         }
	         break;

	      case 2:		// in sync
	         addbittoMP2 (MP2frame, v [i], MP2bitCount ++);
	         if (MP2bitCount >= lf) {
	            processCurrentFrame (lf / 8 + 10);
	            MP2fillState 	= 0;
	            MP2headerCount 	= 0;
	            MP2bitCount		= 0;
	         }
	         break;
	   }
	}
}

void	mp2Processor::addbittoMP2 (uint8_t *v, uint8_t b, int16_t nm) {
uint8_t	byte	= v [nm / 8];
int16_t	bitnr	= 7 - (nm & 7);
uint8_t	newbyte = (01 << bitnr);

	if ((b & 01) == 0)
	   byte	&= ~newbyte;
	else
	   byte |= newbyte;
	v [nm / 8] = byte;
}

void	mp2Processor::setFile (FILE *f) {
	mp2File = f;
}

