#
/*
 *	Once the bits are "in", interpretation and manipulation
 *	should reconstruct the data blocks.
 *	Ofdm_decoder is called once every Ts samples, and results in
 *	2 * Tu bits
 */
#include	"ofdm_decoder.h"
#include	"phasetable.h"
#include	"gui.h"

	ofdm_decoder::ofdm_decoder	(int32_t	T_s,
	                                 int32_t	T_u,
	                                 RingBuffer<DSPCOMPLEX> *iqBuffer,
	                                 bool		lightRun,
	                                 RadioInterface *mr) {
	this	-> T_s			= T_s;
	this	-> T_u			= T_u;
	this	-> iqBuffer		= iqBuffer;
	this	-> lightRun		= lightRun;
	this	-> myRadioInterface	= mr,

	connect (this, SIGNAL (showIQ (int)), mr, SLOT (showIQ (int)));
	this	-> delta	= T_s - T_u;
	phasetable		= new phaseTable ();
	fft_handler		= new common_fft (T_u);
	fft_buffer		= fft_handler -> getVector ();
	phaseReference		= new DSPCOMPLEX [T_u];
	myMapper		= new permVector ();
	iqCount			= 0;
	coarseOffset		= 0;
	displayToken		= 2;
//
	connect (this, SIGNAL (show_snr (int)),
	         mr, SLOT (show_snr (int)));
	snrCount		= 0;
	snr			= 0;
}

	ofdm_decoder::~ofdm_decoder	(void) {
	delete	phasetable;
	delete	fft_handler;
	delete	phaseReference;
	delete	myMapper;
}

//	in practive, we use the "incoming" block
//	and use its data to generate the prs
void	ofdm_decoder::processBlock_0 (DSPCOMPLEX *v) {

	memcpy (fft_buffer, v, T_u * sizeof (DSPCOMPLEX));
	fft_handler	-> do_FFT ();
	fft_handler	-> do_Shift ();
	coarseOffset 	= getMiddle	(fft_buffer);
	snr		=  0.7 * snr + 0.3 * get_snr (fft_buffer);
	if (++snrCount > 10) {
	   show_snr (snr);
	   snrCount = 0;
	}

//	we are now in the frequency domain, and we keep the carriers
//	from -768 .. 768 \{0} in the phaseReference 
	memcpy (phaseReference, fft_buffer, T_u * sizeof (DSPCOMPLEX));
}
//	for the other 75 blocks of data, the first step is to go from
//	time to frequency domain, to get the 1536 carriers.
//	This requires an FFT and zero de-padding and rearrangement
void	ofdm_decoder::processToken (DSPCOMPLEX *inv,
	                            int16_t *ibits, int32_t blkno) {
int16_t		i;
static		int cnt	= 0;
DSPCOMPLEX	*v = (DSPCOMPLEX *)alloca (T_u * sizeof (DSPCOMPLEX));

	memcpy (fft_buffer, &inv [delta], T_u * sizeof (DSPCOMPLEX));
	fft_handler -> do_FFT ();
//	The + and - frequencies are still to be switched, so:
	memcpy (v, &fft_buffer [T_u / 2], T_u / 2 * sizeof (DSPCOMPLEX));
	memcpy (&v [T_u / 2], fft_buffer, T_u / 2 * sizeof (DSPCOMPLEX));
//
//	ok, now we have our 1537 (-768 .. 768) channels back, embedded
//	in -1024 .. 1024 carriers
//
//
//	Differential demodulation means multiplying the
//	incoming symbol with the conjunct of the previous (incoming) symbol
//
//	Logically, differential demodulation is a phase
//	preceding the mapping onto bits as follows
//	for (i = 1024 - 768; i <= 1024 + 768; i ++) {
//	   DSPCOMPLEX x1 = v [i];	// new "previous" symbol
//	   v [i] = v [i] *
//	                         conj (phaseReference [i]);
//	   phaseReference [i]		= x1;
//	}
//
//	Conversion into bits is simple: the two components of the complex
//	each denote a bit. NOTE: x > 0 means 0, x < 0 means 1.
//	Note that myMapper is a function from 0 .. 1535 -> -768 .. 768 \{0}
//	and we are taking the carriers from the shifted fft buffer
//	for (i = 0; i < 1536; i ++) {
//	   DSPCOMPLEX x1 = v [myMapper -> mapIn (i) + 1024];	
//	   DSPFLOAT ab1 = abs (x1);
//	   ibits [i] = real (x1) / ab1 * 256;	// positive = 0, negative = 1
//	   ibits [1536 + i] = imag (x1) / ab1 * 256;
//	}
//
//	The two loops can be combined into a single one
//
	for (i = 0; i < 1536; i ++) {
	   int16_t	index	= myMapper -> mapIn (i) + 1024;
	   DSPCOMPLEX	x = v [index];
	   v [index] = v [index] * conj (phaseReference [index]);
	   phaseReference [index] = x;
	   DSPFLOAT ab1	= abs (v [index]);
// positive = 0, negative = 1
	   ibits [i]	= (real (v [index])) / ab1 * 255.0;
	   ibits [1536 + i] = (imag (v [index])) / ab1 * 255.0;
	}
//
//	and, from time to time we show some clouds
	if (blkno == displayToken) {
	   if (++cnt > 7) {
	      iqBuffer	-> putDataIntoBuffer (&v [1024 - 768], 1536);
	      showIQ	(1536);
	      cnt = 0;
	   }
	}
}

int16_t	ofdm_decoder::coarseCorrector (void) {
	return coarseOffset;
}

int16_t	ofdm_decoder::getMiddle (DSPCOMPLEX *v) {
int16_t		i;
DSPFLOAT	sum = 0;
int16_t		maxIndex = 0;
DSPFLOAT	oldMax	= 0;
//
//	basic sum over 1536 carriers that are - most likely -
//	in the range
	for (i = 150; i < 1536 + 150; i ++)
	   sum += abs (v [i]);
//
//	Now a moving sum, look for a maximum within a reasonable
//	range (around 256, the start of the useful frequencies)
	for (i = 150; i < 350; i ++) {
	   sum -= abs (v [i]);
	   sum += abs (v [i + 1536]);
	   if (sum > oldMax) {
	      sum = oldMax;
	      maxIndex = i;
	   }
	}
	return maxIndex - 255;
}
//

int16_t	ofdm_decoder::get_snr (DSPCOMPLEX *v) {
int16_t	i;
DSPFLOAT	noise 	= 0;
DSPFLOAT	signal	= 0;

	for (i = 20; i < 120; i ++)
	   noise += abs (v [i]);

	for (i = 1024 - 100 - 50; i < 1024 - 100; i ++)
	   signal += abs (v [i]);
	for (i = 1024 + 100 - 50; i < 1024 + 100; i ++)
	   signal += abs (v [i]);

	return get_db (signal / 100) - get_db (noise / 100);
}

