#
/*
 *	We made two implementations of the ofdmProcessor,
 *	one with an explicit "state", the other one with
 *	the state of the computation implicit
 *	This happens to be the explicit one
 */
#include	"ofdm_processor-b.h"
#include	"gui.h"
//
//	States for our little state machine
//
#define	OFDM_INIT		0100
#define	SYNC_ON_NULL		0101
#define	SYNC_ON_END_OF_NULL	0102
#define	OFDM_SYNC_ON_PHASE	0103
#define	OFDM_PRS		0104
#define	OFDM_SYMBOLS		0105

static inline
DSPFLOAT decayingAverage (DSPFLOAT old, DSPFLOAT input, DSPFLOAT weight) {
	if (weight <= 1)
	   return input;
	return input * (1.0 / weight) + old * (1.0 - (1.0 / weight));
}

	ofdm_processor_b::ofdm_processor_b	(virtualInput *theRig,
	                                         uint32_t 	Tnull,
	                                         uint32_t 	T_s,
	                                         uint32_t 	T_u,
	                                         RadioInterface *mr,
	                                         mscHandler 	*msc,
	                                         ficHandler 	*fic,
	                                         int16_t	fillSize,
	                                         RingBuffer<DSPCOMPLEX> *scopeBuffer,
	                                         RingBuffer<DSPCOMPLEX> *iqBuffer,
	                                         int32_t	ibufferSize,
	                                         bool	   lightRun):
	         ofdm_processor (theRig, Tnull, T_s, T_u, mr,
	                         msc, fic, fillSize, scopeBuffer,
	                         iqBuffer, ibufferSize, lightRun) {
	decoderMode	= OFDM_INIT;
	start (QThread::TimeCriticalPriority);
}

	ofdm_processor_b::~ofdm_processor_b	(void) {
	running	= false;
	while (!isFinished ())
	   usleep (100);
}

//
static
int16_t uint8_to_short [256] = {
 -128, -127, -126, -125, -124, -123, -122, -121, -120, -119, -118,
-117, -116, -115, -114, -113, -112, -111, -110, -109, -108, -107,
-106, -105, -104, -103, -102, -101, -100, -99, -98, -97, -96, -95,
-94, -93, -92, -91, -90, -89, -88, -87, -86, -85, -84, -83, -82,
-81, -80, -79, -78, -77, -76, -75, -74, -73, -72, -71, -70, -69,
-68, -67, -66, -65, -64, -63, -62, -61, -60, -59, -58, -57, -56,
-55, -54, -53, -52, -51, -50, -49, -48, -47, -46, -45, -44, -43,
-42, -41, -40, -39, -38, -37, -36, -35, -34, -33, -32, -31, -30,
-29, -28, -27, -26, -25, -24, -23, -22, -21, -20, -19, -18, -17,
-16, -15, -14, -13, -12, -11, -10, -9, -8, -7, -6, -5, -4, -3,
-2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16,
17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32,
33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48,
49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64,
65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80,
81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96,
97, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109,
110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121,
122, 123, 124, 125, 126, 127 };
//
void	ofdm_processor_b::run	 (void) {
int32_t	i, j;
DSPCOMPLEX	Z_Tnull;
DSPCOMPLEX	Z_old;
DSPFLOAT	localLevel;
DSPCOMPLEX	currentValue;
int32_t	sampleCnt	= 0;
int32_t	amount;

	while (running) {
	   volatile int32_t x = theRig -> Samples ();
	   if (x < 10000) {
	      usleep (10);
	      continue;
	   }

	   amount = theRig	-> getSamples (dataBuffer, x);
	   if (dumping)
	      (void)fwrite ((void *)dataBuffer, sizeof (uint8_t),
	                                         2 * amount, dumpFile);
//
	   for (i = 0; i < amount; i ++) {
	      currentValue =
	        DSPCOMPLEX (uint8_to_short [dataBuffer [2 * i]] / 128.0 * gain,
	                    uint8_to_short [dataBuffer [2 * i + 1]] / 128.0 * gain);
//
//	profiling shows that nextValue is an extremely expensive
//	function (due to all checks). Therefore, it is encoded
//	inline
	      localPhase	-=
	            FreqOffset + KHz (f2Corrector) + frequencyCorrector;
	      if (localPhase < 0)
	         localPhase += INPUT_RATE;
	      if (localPhase >= INPUT_RATE)
	         localPhase -= INPUT_RATE;
	      currentValue *= oscillatorTable [localPhase];

	      syncBuffer [syncBufferIndex] = currentValue;
	      Z_Tnull        = 
	               syncBuffer [(syncBufferIndex - Tnull) & syncBufferMask];
	      syncFilterSum	+= abs (currentValue) - abs (Z_Tnull);
	      localLevel 	= syncFilterSum / Tnull;
	      tokenLength ++;
//
//	for the average signal level, we take about 10 times the size of
//	the null symbol
	      signalLevel	+= abs (currentValue) -
	                     abs (syncBuffer [(syncBufferIndex - 10 * Tnull) & syncBufferMask]);
//
//	N times a second we show the spectrum
#define	N	9
	      if (++ sampleCnt > 2048000 / N) {
	         showSpectrum	(fillSize);
	         show_fineCorrector (frequencyCorrector);
	         show_coarseCorrector (f2Corrector);
	         sampleCnt	= 0;
	      }
	
//	To avoid starting up problems, we just let the machine
//	run and fill the buffer to get a more or less reliable
//	signalLevel value
	      switch (decoderMode) {
	         case OFDM_INIT:
	            if (initCount < 10 * Tnull)
	               initCount ++;
	            else {
//	            fprintf (stderr,
//	                     "Moving to sync_on_null %f %f\n",
//	                      signalLevel / (10 * Tnull), abs (currentValue));
	               decoderMode = SYNC_ON_NULL;
	            }
	            break;
//
//	look for a real NULL symbol
	         case SYNC_ON_NULL:
	            if (localLevel < 0.50 * signalLevel / (10 * Tnull)) {
	               nullLevel		= localLevel;
	               decoderMode		= SYNC_ON_END_OF_NULL;
	            }
	            break;

	         case SYNC_ON_END_OF_NULL:
	            if (localLevel > nullLevel) {
	               decoderMode		= OFDM_SYNC_ON_PHASE;
	               syncPhaseIndex		= 1;	// first symbol
	               syncTries		= 0;
	            }
	            else
	              nullLevel			= localLevel;
	            break;

#define	HHH	1536
//	basic synchronization through the phase synchronizer
	         case OFDM_SYNC_ON_PHASE:
	            if (++ syncPhaseIndex >= HHH) {	// we have HHH symbols
	               syncTries += 1;
	               int32_t a = (syncBufferIndex - HHH) & syncBufferMask;
//
//	phaseSynchronizer returns the index of the match, i.e. the
//	first entry of the match
	               int16_t t =
	                     phaseSynchronizer -> findIndex (a, HHH);
	               if (t == -1) {	// not found
	                  if (syncTries > 3)
	                     decoderMode = SYNC_ON_NULL;
	                  break;
	               }
//
//	length of the previous token is to be adjusted:
	               tokenLength		= tokenLength - (HHH - t);
//
//	the tokens for the next token (the PRS symbol) are stored
//	in the ofdmBuffer
	               ofdmBufferIndex = 0;
	               for (j = t; j < HHH; j ++)
//	               for (j = t; j <= HHH; j ++)
	                  ofdmBuffer [ofdmBufferIndex ++] =
	                              syncBuffer [(a + j) & syncBufferMask];
//
//	for experimenting we can set an "offset"
//	from the user interface  
	               if (195000 <= tokenLength &&
	                             tokenLength <= 197500)
	                  avgTokenLength = decayingAverage (avgTokenLength,
	                                                    tokenLength, 3);
	               if (++tokenCount > 10) {
	                  tokenCount = 0;
	                  show_avgTokenLength (avgTokenLength);
	               }
//	... and  prepare for the next round:
	               tokenLength		= HHH - t;
//	and set the Mode
	               decoderMode = OFDM_PRS;
	            }
	            break;
//
//	The synchronisation takes place on (Ts - Tu) samples
//	from the actual beginning, so here we only continue
//	saving samples until Tu.
//	In case the user wants to, we process the PRS symbol and
//	make it into the real reference symbol (using the saved symbols
//	otherwise, we just use a copy of our own PRS symbol)
	         case OFDM_PRS:
	            ofdmBuffer [ofdmBufferIndex ++] = currentValue;
	            if (ofdmBufferIndex >= T_u) {
	               ofdmDecoder	-> processBlock_0 (ofdmBuffer);
	               ofdmBufferIndex	= 0;
	               ofdmSymbolCount	= 1;
	               FreqCorr		= 0;
	               avgCorr		= 0;
	               decoderMode		= OFDM_SYMBOLS;
	            }
	            break;
// 
//	we collect the samples for the ofdm symbol
//	when having Tu samples, we start fine frequency correction
//    	with the cyclic prefix
	         case OFDM_SYMBOLS:
	            ofdmBuffer [ofdmBufferIndex] = currentValue;
	            if (ofdmBufferIndex >= T_u) 
	               FreqCorr += arg (currentValue * 
	                            conj (ofdmBuffer [ofdmBufferIndex - T_u]));

	            if (++ofdmBufferIndex >= T_s) {	// end of block
	               FreqCorr	/=  (2 * M_PI * (T_s - T_u));
	               avgCorr	+= FreqCorr;
	               ++ofdmSymbolCount;
	               ofdmDecoder -> processToken (ofdmBuffer, ibits,
	                                            ofdmSymbolCount);
	               if ((2 <= ofdmSymbolCount) && (ofdmSymbolCount <= 4))
	                     my_ficHandler	-> process_ficBlock (ibits,
	                                                        ofdmSymbolCount);
	               else if (!lightRun)
	                     my_mscHandler	-> process_mscBlock (ibits,
	                                                        ofdmSymbolCount);

	               ofdmBufferIndex = 0;
	               FreqCorr	= 0;
	               if (ofdmSymbolCount >= 76) {	// at the end!!!
	                  fineCorrector	+= (avgCorr / 75) * 1000;
	                  if (fineCorrector > 500) 
	                     fineCorrector -= 500;
	                  if (fineCorrector < -500)
	                     fineCorrector += 500;
	                  frequencyCorrector = 0.9 * frequencyCorrector +
	                                       0.1 * fineCorrector;
	                  coarseCorrector	= ofdmDecoder -> coarseCorrector ();
	                  if (f2Correction)
	                     f2Corrector	+= coarseCorrector;
	                  ofdmSymbolCount	= 0;
	                  decoderMode		= SYNC_ON_NULL;
	               }
	            }
	            break;
   
	         default:;
	      }

	      syncBufferIndex = (syncBufferIndex + 1) & syncBufferMask;
	   }
	}
}
