#
/*
 * 	In order to do some performance testing with various
 * 	coding styles, there are two versions of the ofdm_processor
 * 	One where the "state" is maintained explicitly, i.e. one pushes
 * 	a single sample into the machine, the other where the "state"
 * 	is maintained implicitly, i.e. the running code calls for new
 * 	samples from different places.
 * 	Both are valid implementation, both implementing
 * 	the virtual class ofdm_processor
 */
#
#ifndef	OFDM_DECODER
#define	OFDM_DECODER
/*
 *
 */
#include	"dab-constants.h"
#include	"stdint.h"
#include	"phasereference.h"
#include	"ofdm_decoder.h"
#include	"fic_handler.h"
#include	"msc_handler.h"
#include	"virtual_input.h"
#include	<QThread>
#include	<QObject>
#include	"ringbuffer.h"

class	RadioInterface;
class	common_fft;

class ofdm_processor: public QThread {
Q_OBJECT
public:
		ofdm_processor	(virtualInput *,
	                         uint32_t,
	                         uint32_t,
	                         uint32_t,
	                         RadioInterface *,
	                         mscHandler *,
	                         ficHandler *,
	                         int16_t,		// fillpoint
	                         RingBuffer<DSPCOMPLEX> *,
	                         RingBuffer<DSPCOMPLEX> *,	// passing
	                         int32_t,		// buffersize
	                         bool);
virtual		~ofdm_processor	(void);
	void	setOffset	(int32_t);
	void	setGain		(int16_t);
	void	coarseCorrectorOn	(void);
	void	coarseCorrectorOff	(void);
	void	startDumping	(FILE *);
	void	stopDumping	(void);
	void	showSpectrum	(int16_t);
	bool	running;
	int16_t	gain;
	bool	dumping;
	FILE	*dumpFile;
	int32_t	ibufferSize;
	virtualInput	*theRig;
	RingBuffer<DSPCOMPLEX> * scopeBuffer;
	uint32_t	Tnull;
	uint32_t	T_u;
	uint32_t	T_s;
	RadioInterface	*myRadioInterface;
	bool		lightRun;
	uint8_t		*dataBuffer;
	int32_t		FreqOffset;
	DSPCOMPLEX	*oscillatorTable;
	int32_t		localPhase;
	int16_t		fineCorrector;
	int32_t		coarseCorrector;
	float		f2Corrector;
volatile
	bool		f2Correction;
	int32_t		frequencyCorrector;
	DSPCOMPLEX	*syncBuffer;
	uint32_t	syncBufferSize;
	uint32_t	syncBufferMask;
	uint32_t	syncBufferIndex;
	DSPFLOAT	syncFilterSum;
	DSPFLOAT	signalLevel;
	DSPFLOAT	nullLevel;
	DSPFLOAT	currLevel;
	uint32_t	initCount;
	int32_t		tokenLength;
	int32_t		tokenCount;
	int32_t		avgTokenLength;
	int16_t		*ibits;
	DSPCOMPLEX	*ofdmBuffer;
	uint32_t	ofdmBufferIndex;
	uint32_t	ofdmSymbolCount;
	phaseReference	*phaseSynchronizer;
	ofdm_decoder	*ofdmDecoder;
	DSPFLOAT	FreqCorr;
	DSPFLOAT	avgCorr;
	ficHandler	*my_ficHandler;
	mscHandler	*my_mscHandler;
	int16_t		fillSize;
signals:
	void		show_fineCorrector	(int);
	void		show_coarseCorrector	(int);
	void		show_avgTokenLength	(int);
	void		showScope		(int);
};
#endif

