#
/*
 *
 *    Copyright (C)  2010, 2011, 2012
 *    Jan van Katwijk (J.vanKatwijk@gmail.com)
 *    JFF Consultancy
 *
 *    This file is part of the JSDR.
 *    Many of the ideas as implemented in JSDR are derived from
 *    other work, made available through the GNU general Public License. 
 *    All copyrights of the original authors are recognized.
 *
 *    JSDR is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    JSDR is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with JSDR; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include	"fm_pilotpll.h"
//
	FMPilotPLL::FMPilotPLL (DSPFLOAT gain,
	                        int32_t rate,
	                        int32_t cf,
	                        SinCos *mySinCos) {
	this	-> loopGain		= gain;
	this	-> rate			= rate;
	this	-> pll_cf		= cf;
	this	-> mySinCos		= mySinCos;
	this	-> omega		= (double)pll_cf / rate * (2 * M_PI);
	this	-> locked		= false;
	this	-> pll_corrector	= 0;
	this	-> oldRef		= 0;
	this	-> NcoPhase		= 0;
	this	-> loop_lowPass		= new LowPassIIR (3,
	                                                  350,
	                                                  rate, S_BUTTERWORTH);
	this	-> lock_lowPass		= new LowPassIIR (5,
	                                                  1000,
	                                                  rate, S_BUTTERWORTH);
}

	FMPilotPLL::~FMPilotPLL (void) {
	delete	loop_lowPass;
	delete	lock_lowPass;
}

void	FMPilotPLL::upDate (DSPFLOAT g) {
	loopGain	= g;
}

//
//	Under Fedora we sometimes find signal values inf or nan
DSPFLOAT	FMPilotPLL::process (DSPFLOAT testSignal) {
DSPFLOAT	NcoSignal;
DSPFLOAT	pll_error;
DSPFLOAT	quadRef;
DSPFLOAT	pll_lock;
DSPFLOAT	output;
//
//	first a nan/inf test
	if (isIndeterminate (testSignal) ||
	    isInfinite (testSignal))
	   testSignal = 0;
//
//	from phases to signals:
	NcoSignal = mySinCos -> getCos (NcoPhase);
//
//	phase detector:
	pll_error	= testSignal * NcoSignal * loopGain / 5000;

//	adjust corrector
	pll_corrector	= PI_Constrain (pll_corrector + pll_error);

//	and return the adjusted phase
	output		= PI_Constrain (NcoPhase + pll_corrector);
	NcoPhase	= PI_Constrain (NcoPhase + pll_corrector - omega);
//
//	quadrature reference
	quadRef		= (NcoSignal - oldRef) / omega;
	quadRef		= PI_Constrain (quadRef);
	oldRef		= NcoSignal;
	pll_lock	= lock_lowPass -> Pass (-quadRef * testSignal);
	locked		= pll_lock > 0.10;
	return output;
}

DSPFLOAT	FMPilotPLL::getDelta (void) {
	return pll_corrector;
}

bool		FMPilotPLL::isLocked (void) {
	return locked;
}

