#
/*
 * 	File reader
 */
#include	"filereader.h"
#include	<stdio.h>
#include	<unistd.h>
#include	<stdlib.h>
#include	<fcntl.h>
//

#include	<sys/time.h>
#include	<time.h>

static inline
int64_t		getMyTime	(void) {
struct timeval	tv;

	gettimeofday (&tv, NULL);
	return ((int64_t)tv. tv_sec * 1000000 + (int64_t)tv. tv_usec);
}

#define	INPUT_FRAMEBUFFERSIZE	8 * 32768

	fileReader::fileReader (const char *f, bool *success) {
	fileName	= f;
	readerOK	= false;
	_I_Buffer	= new RingBuffer<uint8_t>(INPUT_FRAMEBUFFERSIZE);
	filePointer	= fopen (f, "r");
	if (filePointer == NULL) {
	   fprintf (stderr, "file %s cannot open\n", f);
	   *success = false;
	   return;
	}
	readerOK	= true;
	readerPausing	= true;
	*success	= true;
	currPos		= 0;
	start	();
}

	fileReader::~fileReader (void) {
	ExitCondition = true;
	if (readerOK) {
	   while (isRunning ())
	      usleep (100);
	   fclose (filePointer);
	}
	delete _I_Buffer;
}

bool	fileReader::restartReader	(void) {
	if (readerOK)
	   readerPausing = false;
	return readerOK;
}

void	fileReader::stopReader		(void) {
	if (readerOK)
	   readerPausing = true;
}

uint8_t	fileReader::myIdentity		(void) {
	return FILEREADER;
}

int32_t	fileReader::getSamples		(uint8_t *out, int32_t amount) {
int32_t	count;

	if (filePointer == NULL)
	   return 0;

	while (_I_Buffer -> GetRingBufferReadAvailable () < 2 * amount)
	   msleep (1);

	_I_Buffer	-> getDataFromBuffer (out, 2 * amount);
	return amount;
}

int32_t	fileReader::Samples (void) {
	return _I_Buffer -> GetRingBufferReadAvailable () / 2;
}


void	fileReader::run (void) {
int32_t	t, i;
uint8_t	*bi;
int32_t	bufferSize	= 32768;
int64_t	period;
int64_t	nextStop;

	if (!readerOK)
	   return;

	ExitCondition = FALSE;
	ThreadFinished	= FALSE;

	period		= (32768 * 1000) / (2 * 2048);	
	fprintf (stderr, "Period = %d\n", period);
	bi		= new uint8_t [bufferSize];
	nextStop	= getMyTime ();
	while (!ExitCondition) {
	   if (readerPausing) {
	      usleep (1000);
	      nextStop = getMyTime ();
	      continue;
	   }
	   while (_I_Buffer -> WriteSpace () < bufferSize + 10) {
	      if (ExitCondition)
	         break;
	      usleep (100);
	   }

	   nextStop += period;
	   t = readBuffer (bi, bufferSize);
	   if (t <= 0) {
	      for (i = 0; i < bufferSize; i ++)
	          bi [i] = 0;
	      t = bufferSize;
	   }
	   _I_Buffer -> putDataIntoBuffer (bi, t);
	   if (nextStop - getMyTime () > 0)
	      usleep (nextStop - getMyTime ());
	}
	fprintf (stderr, "taak voor replay eindigt hier\n");
	ThreadFinished	= TRUE;
}
/*
 *	length is number of uints that we read.
 */
int32_t	fileReader::readBuffer (uint8_t *data, int32_t length) {
int32_t	n, i;

	n = fread (data, sizeof (uint8_t), length, filePointer);
	currPos		+= n;
	if (n < length) {
	   fseek (filePointer, 0, SEEK_SET);
	   fprintf (stderr, "End of file, restarting\n");
	}
	return	n & ~01;
}

