#
/*
 *    Copyright (C) 2010,2011, 2012
 *    Jan van Katwijk (J.vanKatwijk@gmail.com)
 *    JFF Consultancy
 *
 *    This file is part of the  JSDR (JSDR).
 *    Many of the ideas as implemented in JSDR are derived from
 *    other work, made available through the GNU general Public License. 
 *    All copyrights of the original authors are recognized.
 *
 *    JSDR is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    JSDR is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with JSDR; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include	<QSettings>
#include	<QMessageBox>
#include	<QFileDialog>
#include	<QDebug>
#include	<qwt_color_map.h>
#include	<qwt_plot_spectrogram.h>
#include	<qwt_scale_widget.h>
#include	<qwt_scale_draw.h>
#include	<qwt_plot_zoomer.h>
#include	<qwt_plot_panner.h>
#include	<qwt_plot_layout.h>
#include	<QDateTime>
#include	<QFile>
#include	"gui.h"
#ifdef	__MINGW32__
#include	"dabstick-dll.h"
#else
#include	"dabstick.h"
#endif
#include	"scope.h"
#ifdef __MINGW32__
#include	<iostream>
#endif
#
//
#include	<QKeyEvent>

#define	SCAN_DELAY	300
#define	IDLE		0100
#define	PAUSED		0101
#define	RUNNING		0102

//
/*
 *	We use the creation function merely to set up the
 *	user interface and make the connections between the
 *	gui elements and the handling agents. All real action
 *	is embedded in actions, initiated by gui buttons
 */
	RadioInterface::RadioInterface (QSettings	*Si,
	                                QWidget		*parent): QDialog (parent) {
bool	success;
int	i;
// 	the setup for the generated part of the ui
	setupUi (this);
	ISettings			= Si;
	this		-> doInit	= doInit;
	this		-> CardRate 	= 2500000;
	this		-> segmentSize	= 512;
	this		-> segmentCount	= 5;
	this		-> rasterSize	= 50;
	this		-> spectrumFactor	= 8;
	this		-> scanDelay	= SCAN_DELAY;
	currentFrequency		= 100000000;
	if (ISettings != NULL) {
	   this -> CardRate  
	           = ISettings -> value ("CardRate", 2500000). toInt ();

	   this -> segmentSize =
	             ISettings -> value ("segmentSize", 256). toInt ();
	   this -> segmentCount =
	             ISettings -> value ("segmentCount", 5). toInt ();
	   this -> spectrumFactor = 
	             ISettings -> value ("spectrumFactor", 8). toInt ();
	   this -> scanDelay	=
	             ISettings -> value ("scanDelay", SCAN_DELAY). toInt ();
	   currentFrequency = Khz (
	              ISettings -> value ("currentFrequency", 100000). toInt ());
	}

	if ((segmentSize & (segmentSize - 1)) != 0)
	   spectrumSize = 128;

	lcd_segments	-> display (segmentCount);
	lcd_Width	-> display (segmentCount * CardRate);
	lcd_InputRate	-> display ((int)this -> CardRate);
	displaySize	= segmentCount * segmentSize;
	spectrumSize	= spectrumFactor * segmentSize;
	this		-> newSegment	= true;
	this		-> currentSegment	= 0;
	Window				= new DSPFLOAT [spectrumSize / 2];
	for (i = 0; i < spectrumSize / 2; i ++)
	   Window [i] = 0.42 - 0.5 * cos ((2.0 * M_PI * i) / (spectrumSize / 2)) +
	                      0.08 * cos ((4.0 * M_PI * i) / (spectrumSize / 2));


	HFScope		= new Scope (hfscope,
	                             this -> displaySize,
	                             this -> rasterSize);
	HFViewmode	= SPECTRUM_MODE;
	HFScope		-> SelectView (SPECTRUM_MODE);
	connect (HFScope, SIGNAL (clickedwithLeft (int)),
	         this, SLOT (AdjustFrequency (int)));
	this		-> X_axis		= new double [displaySize];
	this		-> displayVector	= new double [displaySize];

	displayBuffer				= new double *[segmentCount];
	for (i = 0; i < segmentCount; i ++)
	   displayBuffer [i] = new double [segmentSize];
	this		-> spectrum_fft	= new common_fft (spectrumSize);
	this		-> spectrumBuffer	= spectrum_fft -> getVector ();
//
//	set some sliders to their values
	if (ISettings != NULL) {
	int k;
	   k	= ISettings -> value ("spectrumAplitudeSlider", 20). toInt ();
	   spectrumAmplitudeSlider	-> setValue (k);
	   k	= ISettings -> value ("attenuationSlider", 20). toInt ();
	   attenuationSlider		-> setValue (k);
	   k	= ISettings -> value ("IQbalanceSlider", 0). toInt ();
	   IQbalanceSlider		-> setValue (k);
	}

	ClearPanel		();
	connect (startButton, SIGNAL (clicked ()),
	              this, SLOT (setStart ()));
	connect (QuitButton, SIGNAL (clicked ()),
	              this, SLOT (TerminateProcess ()));
	connect (viewmodeButton, SIGNAL (clicked (void)),
	              this, SLOT (setViewmode (void)));
	connect (spectrumAmplitudeSlider, SIGNAL (valueChanged (int)),
	         this,  SLOT (setAmplification (int)));
	connect (attenuationSlider, SIGNAL (valueChanged (int) ),
		      this, SLOT (setAttenuation (int) ) );
	connect (IQbalanceSlider, SIGNAL (valueChanged (int)),
	              this, SLOT (setIQBalance (int)));
	connect (averagerButton, SIGNAL (clicked (void)),
	              this, SLOT (setAverager (void)));
/*
 *	connections for the console, first the digits
 */
	connect (add_one,   SIGNAL (clicked() ), this, SLOT (addOne() ) );
	connect (add_two,   SIGNAL (clicked() ), this, SLOT (addTwo() ) );
	connect (add_three, SIGNAL (clicked() ), this, SLOT (addThree() ) );
	connect (add_four,  SIGNAL (clicked() ), this, SLOT (addFour() ) );
	connect (add_five,  SIGNAL (clicked() ), this, SLOT (addFive() ) );
	connect (add_six,   SIGNAL (clicked() ), this, SLOT (addSix() ) );
	connect (add_seven, SIGNAL (clicked() ), this, SLOT (addSeven() ) );
	connect (add_eight, SIGNAL (clicked() ), this, SLOT (addEight() ) );
	connect (add_nine,  SIGNAL (clicked() ), this, SLOT (addNine() ) );
	connect (add_zero,  SIGNAL (clicked() ), this, SLOT (addZero() ) );
/*
 *	function buttons
 */
	connect (khzSelector, SIGNAL (clicked ()),
	                      this, SLOT (AcceptFreqinKhz ()));
	connect (add_correct, SIGNAL (clicked() ), this, SLOT (addCorr() ) );
	connect (add_clear,   SIGNAL (clicked() ), this, SLOT (addClear() ) );

	connect (pauseButton, SIGNAL (clicked (void)),
	         this, SLOT (clickPause (void)));

#ifdef	__MINGW32__
	MyRig	= new dabstick_dll (this -> CardRate, 4, 1024, &success);
#else
	MyRig	= new dabstick (this -> CardRate, 4, 1024,  &success);
#endif
	
	if (!success) {
	   QMessageBox::warning (this, tr ("sdr"),
	                               tr ("Could not open dongle"));
	   delete MyRig;
	   exit (101);
	}

	fprintf (stderr, "found the rig\n");
//
//	Create a timer for dealing with a non acting (or reacting) user
//	on the keyboard
	lcd_timer		= new QTimer ();
	lcd_timer		-> setSingleShot (TRUE);	// single shot
	lcd_timer		-> setInterval (5000);
	connect (lcd_timer, SIGNAL (timeout ()),
	         this, SLOT (lcd_timeout ()));

	runTimer		= new QTimer ();
	runTimer		-> setSingleShot (FALSE);
	runTimer		-> setInterval (50);	// milleseconds

	connect (runTimer, SIGNAL (timeout (void)),
	         this, SLOT (JansRadioThroughSignal (void)));

	scanTimer		= new QTimer ();
	scanTimer		-> setSingleShot (TRUE);
	scanTimer		-> setInterval (scanDelay);
	connect (scanTimer, SIGNAL (timeout (void)),
	         this, SLOT (updateScanTimer (void)));
/*
 *	just as a gadget, we show the actual time
 */
	display_timer		= new QTimer ();
	display_timer		-> setInterval (1000);
	connect (display_timer,
	         SIGNAL (timeout ()),
	         this,
	         SLOT (updateTimeDisplay ()));

	char	*s = (char *)alloca (256);
	sprintf (s, "sdrJ-SB-V%s", CURRENT_VERSION);
	systemindicator		-> setText (QString (s));
	display_timer		-> start (1000);
//
//	all elements seem to exist: set the control state to
//	its default values
	setAmplification	(spectrumAmplitudeSlider -> value ());
	setAttenuation		(attenuationSlider	 -> value ());
	setIQBalance		(IQbalanceSlider	-> value ());
	runMode			= IDLE;
	averager		= 0;
}

	RadioInterface::~RadioInterface (void) {
	if (ISettings != NULL) {
	   int temp;
	   ISettings -> setValue ("CardRate", CardRate);
	   ISettings -> setValue ("segmentSize", segmentSize);
	   ISettings -> setValue ("segmentCount", segmentCount);
	   ISettings -> setValue ("spectrumFactor", spectrumFactor);
	   ISettings -> setValue ("scanDelay", scanDelay);
	   temp = (int)(currentFrequency / Khz (1));
	   ISettings -> setValue ("currentFrequency", temp);
	   ISettings -> setValue ("spectrumAplitudeSlider",
	                           spectrumAmplitudeSlider -> value ());
	   ISettings -> setValue ("attenuationSlider",
	                           attenuationSlider -> value ());
	   ISettings -> setValue ("IQbalanceSlider",
	                           IQbalanceSlider -> value ());
	}

	delete		lcd_timer;
	delete		MyRig;
	delete		HFScope;
	delete		display_timer;
	delete		scanTimer;
	if (Window != NULL)
	   delete[]	Window;
}
//
//	Now start only means anything if we are not running already
void	RadioInterface::setStart	(void) {
bool	r = 0;

	if (runMode != IDLE)
	   return;

	Set_Tuner (currentFrequency);
	Display (currentFrequency, 0);
	MyRig		-> stopReader ();	
//
	r = MyRig -> restartReader ();
	qDebug ("Starting %d\n", r);
	if (!r) {
	   QMessageBox::warning (this, tr ("sdr"),
	                               tr ("Opening  input stream failed\n"));
	   return;
	}

//	and we are on the move, so let the timer run
	runTimer	-> start (50);
	runMode		= RUNNING;
}

void	RadioInterface::TerminateProcess () {
//	stop the usb handler!!
	MyRig 		-> stopReader ();
	runTimer	-> stop ();
	accept ();
	qDebug () <<  "End of termination procedure";
}

void	RadioInterface::abortSystem (int d) {
	qDebug ("aborting for reason %d\n", d);
	exit (11);
	accept ();
	
}

void	RadioInterface::setAmplification (int a) {
	scopeAmplifier = a;
}

void	RadioInterface::setAttenuation (int n) {
int16_t	f	= IQBalanceDisplay	-> value ();
int16_t	bl, br;

	bl	= 100 - 2 * f;
	br	= 100 + 2 * f;
	attenuationLevelDisplay	-> display (n);
	currAttSliderValue	= n / 10.0;
	attValueL	= currAttSliderValue * (float)bl / 100;
	attValueR	= currAttSliderValue * (float)br / 100;
}
/*
 * 	the balance slider runs between -30 .. 30
 */
void	RadioInterface::setIQBalance (int n) {
int16_t	bl, br;

	IQBalanceDisplay -> display (n);
	bl		= 100 - 2 * n;
	br		= 100 + 2 * n;
	attValueL	= currAttSliderValue * (float)bl / 100;
	attValueR	= currAttSliderValue * (float)br / 100;
}
/*
 * 	Handling the numeric keypad is boring, but needed
 */
void RadioInterface::addOne (void) {
	AddtoPanel (1);
}

void RadioInterface::addTwo (void) {
	AddtoPanel (2);
}

void RadioInterface::addThree (void) {
	AddtoPanel (3);
}

void RadioInterface::addFour (void) {
	AddtoPanel (4);
}

void RadioInterface::addFive (void) {
	AddtoPanel (5);
}

void RadioInterface::addSix (void) {
	AddtoPanel (6);
}

void RadioInterface::addSeven (void) {
	AddtoPanel (7);
}

void RadioInterface::addEight (void) {
	AddtoPanel (8);
}

void RadioInterface::addNine (void) {
	AddtoPanel (9);
}

void RadioInterface::addZero (void) {
	AddtoPanel (0);
}
//
//	If we are keying, the lcd timer is on, so
//	if it is not on, we are not keying and clear does not
//	mean anything
void RadioInterface::addClear (void) {
	if (!lcd_timer -> isActive ())
	   return;

	stop_lcdTimer	();
	if (scanTimer	-> isActive ())
	   scanTimer	-> stop ();

	ClearPanel ();
	Display (currentFrequency, 0);
}
//
//	Note that in fm receiver, user commands are in KHz
//
void RadioInterface::AcceptFreqinKhz (void) {
int32_t	p;

	if (!lcd_timer -> isActive ())
	   return;

	stop_lcdTimer	();
	if (scanTimer	-> isActive ())
	   scanTimer	-> stop ();
//	we know that the increment timer is not running,
//	so there is no reason to stop it
	p = getPanel ();
	ClearPanel ();
	Set_Tuner (Khz (p));
	Display (currentFrequency, 0);
}

void RadioInterface::addCorr (void) {
	if (!lcd_timer -> isActive ()) 
	   return;

	stop_lcdTimer	();
//	increment timer cannot run now so no reason
//	to stop it
	Panel = (Panel - (Panel % 10)) / 10;
	Display	(Panel, 1);
	lcd_timer	-> start (5000);		// restart timer
}
//
//	In AddtoPanel, we might have any of the two
//	timers running, do not know which one though
void	RadioInterface::AddtoPanel (int16_t n) {
	stop_lcdTimer	();
	if (scanTimer -> isActive ())
	   scanTimer -> stop ();
	Panel		= 10 * Panel + n;
	Display (Panel, 1);
	lcd_timer	-> start (5000);		// restart timer
}

void	RadioInterface::ClearPanel (void) {
	Panel = 0;
}

int	RadioInterface::getPanel (void) {
	return Panel;
}

void	RadioInterface::AdjustFrequency (int n) {
	stop_lcdTimer ();
	if (scanTimer -> isActive ())
	   scanTimer -> stop ();

	Set_Tuner (currentFrequency + Khz (n));
	Display (currentFrequency, 0);
}

//	Set_Tuner accepts freq in Hz.
//
void	RadioInterface::Set_Tuner (int32_t n) {
double	temp	= (double)(segmentCount * CardRate) / 2 / displaySize;
int32_t	i;

//	first, while we are building an image, just stop it
	if (scanTimer -> isActive ()) 
	   scanTimer -> stop ();
//
//	prepare, set up for new frames
	frequency	= n - (segmentCount - 1) * CardRate / 2;
	currentSegment	= 0;
	MyRig		-> setVFOFrequency (frequency);
#ifdef	__MINGW32__
	((dabstick_dll *)MyRig)		-> resetBuffer ();
#else
	((dabstick *)MyRig)		-> resetBuffer ();
#endif
//
	currentFrequency	= n;
//	a new frequency implies a new X_axis
	for (i = 0; i < displaySize; i ++) {
	   X_axis [i] = (currentFrequency - segmentCount * CardRate / 2 +
	                ((double)i) * 2.0 * temp) / ((double)1000);
	   displayVector [i] = 0;
	   if (averager != 0)
	      averager = 2;
	}
	newSegment	= true;
	scanTimer	-> start (scanDelay);
//
//	and start the timer
}
//
//	The clock will drive the updates
void	RadioInterface::updateScanTimer (void) {
	if (++ currentSegment >= segmentCount) {	// at the end
	   currentSegment = 0;
	   frequency = currentFrequency - (segmentCount - 1) * CardRate / 2;
//	and since we are increasing the frequency:
	}
	else
	   frequency += CardRate;

	MyRig	-> setVFOFrequency (frequency);
#ifdef	__MINGW32__
	((dabstick_dll *)MyRig)		-> resetBuffer ();
#else
	((dabstick *)MyRig)		-> resetBuffer ();
#endif
	scanTimer -> start (scanDelay);
	newSegment = true;
}

void	RadioInterface::stop_lcdTimer (void) {
	if (lcd_timer -> isActive ())
	   lcd_timer -> stop ();
}

void	RadioInterface::updateTimeDisplay (void) {
QDateTime	currentTime = QDateTime::currentDateTime ();

	timeDisplay	-> setText (currentTime.
	                            toString (QString ("dd.MM.yy:hh:mm:ss")));
}

void	RadioInterface::lcd_timeout (void) {
	Panel		= 0;			// throw away anything
	Display (currentFrequency, 0);
}

void	RadioInterface::clickPause (void) {
	if (runMode == IDLE)
	   return;

	if (runMode == RUNNING) {
	   MyRig	-> stopReader ();
	   pauseButton	-> setText (QString ("Continue"));
	   if (runTimer -> isActive ())
	      runTimer -> stop ();
	   runMode = PAUSED;
	   return;
	}
	if (runMode == PAUSED) {
	   MyRig	-> restartReader ();
	   pauseButton	-> setText (QString ("Pause"));
	   Set_Tuner (currentFrequency);
	   Display (currentFrequency, 0);
//	   and we are on the move, so let the timer run
	   runTimer	-> start (50);
	   runMode = RUNNING;
	   return;
	}
//
//	We should not be here
}
	
//
DSPFLOAT	decayingAverage (DSPFLOAT old, DSPFLOAT input, DSPFLOAT weight) {
	if (weight <= 1)
	   return input;
	return input * (1.0 / weight) + old * (1.0 - (1.0 / weight));
}
//
//	we want N times per second a signal to start the function
//	we want to process CardRate / 20 length  segments
void	RadioInterface::JansRadioThroughSignal (void) {
uint8_t	*v = (uint8_t *)alloca (spectrumSize / 2 * 2 *  sizeof (uint8_t));
double	*Vector = (double *)alloca (displaySize * sizeof (double));
int32_t		i, j;
int32_t		currentAmount = MyRig -> Samples ();
static	int	renewCount	= 0;
static
double	*currentVector	= displayBuffer [0];

	if (currentAmount < CardRate / 20)
	   return;

	MyRig -> getSamples (v, spectrumSize / 2, CardRate / 20);

	if (currentSegment < 0)	// should not happen, just ignore input
	   return;

	if (newSegment) {
	   newSegment = false;
	   currentVector	= displayBuffer [currentSegment];
	   for (i = 0; i < segmentSize; i ++)
	      currentVector [i] = 0;
	}
//	
	MaptoSegment (v, spectrumSize / 2, currentVector, segmentSize);

	if (++renewCount > 1) {
	   renewCount = 0;
	   for (i = 0; i < segmentCount; i ++)
	      for (j = 0; j < segmentSize; j ++)
	         displayVector [i * segmentSize + j] =
	                 decayingAverage (displayVector [i * segmentSize + j],
	                                  displayBuffer [i][j],
	                                  float(averager));
	   if (averager != 0) averager ++;

	   memcpy (Vector, displayVector, displaySize * sizeof (double));
	   HFScope -> Display (X_axis, 
	                       Vector,
	                       scopeAmplifier,
	                       currentFrequency / 1000);
	}
}
//
//	Note: amount indicates the number of pairs
void	RadioInterface::MaptoSegment (uint8_t *v, int32_t amount,
	                              double *segment, int32_t segmentSize) {
int32_t	i, j;

	for (i = 0; i < amount; i ++)
	   spectrumBuffer [i] =
	          DSPCOMPLEX ((v [2 * i] - 127.0) / 128.0  * attValueL * Window [i],
	                      (v [ 2 * i + 1] - 127.0) / 128.0 * attValueR * Window [i]); 
	for (i = amount; i < spectrumSize; i ++)
	   spectrumBuffer [i] = 0;
	spectrum_fft	-> do_FFT ();
//	we know that spectrumSize = K * segmentSize;
	for (i = 0; i < segmentSize / 2; i ++) {
	   segment [i] = 0;
	   segment [segmentSize / 2 + i] = 0;
	   for (j = 0; j < spectrumSize / segmentSize; j ++) {
	      segment [i] += abs (spectrumBuffer [spectrumSize / 2 + i * spectrumSize / segmentSize + j]);
	      segment [segmentSize / 2 + i] +=
	                     abs (spectrumBuffer [i * spectrumSize / segmentSize + j]);
	   }

	   segment [i] /= spectrumSize / segmentSize;
	   segment [segmentSize / 2 + i ] /= spectrumSize / segmentSize;
	}
}
	
static inline
int32_t numberofDigits (int32_t f) {

	if (f < 100000)
	   return 6;
	if (f < 100000000)
	   return 8;
	if (f < 1000000000)
	   return 9;
	else
	   return 10;
}

void	RadioInterface::Display (int32_t freq, int16_t Mode) {
int32_t nd	= numberofDigits (freq);

	if (Mode == 0) {
	   lcd_Frequency	-> setNumDigits (nd);
	   lcd_Frequency	-> display ((int)freq / 1000);
	}
	else {
	   lcd_Frequency	-> setNumDigits (nd);
	   lcd_Frequency	-> display ((int)freq);
	}
}

void	RadioInterface::setViewmode (void) {
	if (HFViewmode == SPECTRUM_MODE)
	   HFViewmode = WATERFALL_MODE;
	else
	   HFViewmode = SPECTRUM_MODE;

	HFScope -> SelectView (HFViewmode);
}

void	RadioInterface::setAverager	(void) {
int32_t	i;

	if (averager == 0) {
	   averager = 2;
	   averagerButton	-> setText ("Averaging On");
	}
	else {
	   averager = 0;
	   averagerButton	-> setText ("Averager");
	}
}

