#
/*
 *
 *    Copyright (C)  2010, 2011
 *    Jan van Katwijk (J.vanKatwijk@gmail.com)
 *    JFF Consultancy
 *
 *    This file is part of the JFF SDR (JSDR).
 *    Many of the ideas as implemented in JSDR are derived from
 *    other work, made available through the GNU general Public License. 
 *    All copyrights of the original authors are recognized.
 *
 *    JSDR is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    JSDR is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with JSDR; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *	This pll was found to give reasonable results.
 *	source DTTSP, all rights acknowledged
 *	
 *	This is the version for float/double values, which
 *	is slightly different from the complex one
 */
#include	"pllR.h"
//#include	"utilities.h"
//
//	rate	is the samplerate
//	lofreq and hifreq the frequencies (in Hz) where the lock is 
//	kept in between,
//	bandwidth the bandwidth of the signal to be received
//	the SinCos is a class providing table-based access to
//	sine, cosine and complex values.

		pllR::pllR (int32_t rate,
	                    DSPFLOAT freq,
	                    DSPFLOAT lofreq, DSPFLOAT hifreq,
	                    DSPFLOAT bandwidth,
	                    SinCos *Table) {
DSPFLOAT	fac	= 2.0 * M_PI / rate;

	omega		= freq * fac;
	NcoPhaseIncr	= freq * fac ;
	NcoLLimit	= lofreq * fac;
	NcoHLimit	= hifreq * fac;
	pll_Alpha	= 0.125 * bandwidth * fac;	// pll bandwidth
	pll_Beta	= (pll_Alpha * pll_Alpha) / 4.0; // second order term
	NcoPhase	= 0;
	mySinCos	= Table;
	phzError	= 0;
	pll_lock	= 0;
}

		pllR::~pllR (void) {
}
//
//	Under Fedora we found sometimes a signal value of
//	infinite. We therefore constrain the value
void		pllR::do_pll (DSPFLOAT signal) {
DSPFLOAT	NcoSignal;
DSPFLOAT	quadRef;

	if (signal < -100 || signal > 100)
	   signal = 0;

	NcoSignal = (mySinCos != NULL) ?
	                mySinCos -> getCos (NcoPhase):
	                cos (NcoPhase);

	phzError  	= signal * NcoSignal;
	NcoPhaseIncr	+= pll_Beta * phzError;
	if (NcoPhaseIncr < NcoLLimit)
	   NcoPhaseIncr = NcoLLimit;
	if (NcoPhaseIncr > NcoHLimit)
	   NcoPhaseIncr = NcoHLimit;

	NcoPhase	+= NcoPhaseIncr + pll_Alpha * phzError;
	if (NcoPhase >= 2 * M_PI)
	   NcoPhase = fmod (NcoPhase, 2 * M_PI);
	while (NcoPhase < 0)
	   NcoPhase += 2 * M_PI;

	quadRef		= (NcoSignal - oldNcoSignal) / omega;
//	quadRef		= PI_Constrain (quadRef);
	oldNcoSignal	= NcoSignal;
	pll_lock	= 1.0 / 30 * (-quadRef * signal) +
	                  (1.0 - (1.0 / 30)) * pll_lock;
}

DSPFLOAT	pllR::getPhaseIncr(void) {
	return NcoPhaseIncr;
}

DSPFLOAT	pllR::getNco (void) {
	return NcoPhase;
}

bool		pllR::isLocked	(void) {
	return pll_lock > 0.10;
}

DSPFLOAT	pllR::getPhaseError (void) {
	return phzError;
}

