#
/*
 *
 *    Copyright (C) 2008, 2009, 2010
 *    Jan van Katwijk (J.vanKatwijk@gmail.com)
 *    JFF Consultancy
 *
 *    This file is part of the JFF SDR (JSDR).
 *    Many of the ideas as implemented in JSDR are derived from
 *    other work, made available through the GNU general Public License. 
 *    All copyrights of the original authors are recognized.
 *
 *    JSDR is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    JSDR is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with JSDR; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include	"sincos.h"
//
//	As it turns out, when using DAB sticks, this simple function is the
//	real CPU burner, with a usage of up to 22 %

	SinCos::SinCos (DSPCOMPLEX *Table, int32_t Rate) {
           this	-> Table	= Table;
	   this	-> localTable	= false;
	   this	-> Rate		= Rate;
}

	SinCos::SinCos (int32_t Rate) {
int32_t	i;
	   this	-> Rate		= Rate;
	   this	-> localTable	= true;
	   this	-> Table	= new DSPCOMPLEX [Rate];
	   for (i = 0; i < Rate; i ++) 
	      Table [i] = DSPCOMPLEX (cos (2 * M_PI * i / Rate),
	                              sin (2 * M_PI * i / Rate));
}

	SinCos::~SinCos (void) {
	   if (localTable)
	      delete [] Table;
}
//	Heavy code: executed millions of times
//	we get all kinds of very strange values here, so
//	testing over the whole domain is needed
static inline
int32_t	fromPhasetoIndex (DSPFLOAT Phase, int32_t Rate) {
	if (0 <= Phase && Phase < 2 * M_PI)
	   return Phase / (2 * M_PI) * Rate;

	if (Phase >= 2 * M_PI)
	   return fmod (Phase, 2 * M_PI) / (2 * M_PI) * Rate;

	if (Phase >= -2 * M_PI) {
	   Phase = -Phase;
	   return Rate - Phase / (2 * M_PI) * Rate;
	}

	Phase = -Phase;
	return Rate - fmod (Phase, 2 * M_PI) / (2 * M_PI) * Rate;
}

DSPFLOAT	SinCos::getSin (DSPFLOAT Phase) {
	return imag (Table [fromPhasetoIndex (Phase, Rate)]);
}

DSPFLOAT	SinCos::getCos (DSPFLOAT Phase) {
	return real (Table [fromPhasetoIndex (Phase, Rate)]);
}

DSPCOMPLEX	SinCos::getComplex (DSPFLOAT Phase) {
	return Table [fromPhasetoIndex (Phase, Rate)];
}


